from just_the_facts.rollout import FactsScenario

train_urls = [
    "https://www.nbcnews.com/select/shopping/best-dusters-rcna225656",
    "https://www.foxnews.com/us/us-military-raid-syria-eliminates-isis-leader-waiting-key-financier-officials",
    "https://www.bbc.com/news/articles/cgqnzy9ej0do?at_medium=RSS&at_campaign=rss",
    "https://www.nbcnews.com/business/economy/trump-lisa-cook-federal-reserve-feud-flares-up-rcna226049",
    "https://www.foxnews.com/politics/mamdani-sidesteps-communism-question-says-his-campaign-about-distribution-wealth",
    "https://www.nbcnews.com/news/us-news/video-shows-colombian-tiktoker-tatiana-martinez-yanked-car-ice-arrest-rcna226038",
    "https://www.foxnews.com/politics/newsom-style-redistricting-efforts-critiqued-california-democrats-recently-july-statements-show",
    "https://www.nbcnews.com/weather/hurricanes/hurricane-erin-prompts-tropical-storm-warning-north-carolina-rcna226017",
    "https://www.cbsnews.com/news/google-pixel-10-product-launch-jimmy-fallon-gemini/",
    "https://www.cbsnews.com/news/trump-lisa-cook-federal-reserve-governor-bill-pulte/",
    "https://www.foxnews.com/politics/vance-says-national-guard-busting-a-washington-floats-mission-extension",
    "https://www.bbc.com/news/articles/c1ejvxne7elo?at_medium=RSS&at_campaign=rss",
    "https://www.foxnews.com/world/ukraines-stolen-children-crisis-looms-large-nato-meets-russias-war",
    "https://www.bbc.com/news/articles/clyr7l0z9edo?at_medium=RSS&at_campaign=rss",
    "https://www.bbc.com/news/articles/cly4v3pllgko?at_medium=RSS&at_campaign=rss",
    "https://www.bbc.com/news/articles/c24zdel5j18o?at_medium=RSS&at_campaign=rss",
    "https://www.nbcnews.com/news/obituaries/tiktok-star-john-crawley-better-known-kingbeardx-dies-47-rcna226055",
    "https://www.cbsnews.com/news/ten-commandments-texas-abbott-law-blocked-federal-judge/",
    "https://www.bbc.com/news/articles/cd0d31r423ko?at_medium=RSS&at_campaign=rss",
    "https://www.nbcnews.com/select/shopping/hand-soap-recall-rcna224873",
    "https://www.foxnews.com/us/family-torn-georgia-teen-accused-killing-parents-denied-bond-emotional-hearing",
    "https://www.foxnews.com/media/chris-matthews-tells-dems-abandon-all-other-crap-open-borders-trans-issues-focus-economy",
    "https://www.bbc.com/news/articles/c3r413l5n57o?at_medium=RSS&at_campaign=rss",
    "https://www.bbc.com/news/articles/cj0yd0829m4o?at_medium=RSS&at_campaign=rss",
    "https://www.nbcnews.com/business/markets/trump-bought-bonds-companies-governments-rcna226024",
    "https://www.bbc.com/news/articles/c4g64qedynlo?at_medium=RSS&at_campaign=rss",
    "https://www.nbcnews.com/politics/justice-department/judge-denies-doj-request-release-jeffrey-epstein-grand-jury-transcript-rcna226117",
    "https://www.cbsnews.com/news/hurricane-erin-track-path/",
    "https://www.foxnews.com/politics/duffys-dot-accuses-biden-buttigieg-inflating-air-traffic-controller-pipeline-juiced-numbers",
    "https://www.cbsnews.com/news/dhs-secretary-says-entire-southern-border-wall-to-be-painted-black-to-stop-people-from-climbing-it/",
    "https://www.bbc.com/news/articles/cy08309kxw6o?at_medium=RSS&at_campaign=rss",
    "https://www.cbsnews.com/news/us-deportation-agreements-honduras-uganda/",
    "https://www.bbc.com/news/articles/c776kynn771o?at_medium=RSS&at_campaign=rss",
    "https://www.foxnews.com/politics/vance-white-house-blast-crazy-communists-protesting-dc-clean-uo-terrorizing-locals-stupid-white-hippies",
    "https://www.foxnews.com/health/jeff-bezos-moms-death-spotlights-lewy-body-dementia-symptoms-risks",
    "https://www.bbc.com/news/articles/cy8jlwee0xjo?at_medium=RSS&at_campaign=rss",
    "https://www.foxnews.com/politics/letitia-james-lawyer-blasts-ed-martins-bizarre-stunt-new-york-ags-house",
    "https://www.foxnews.com/politics/hillary-clinton-slapped-with-ethics-complaint-targeting-her-law-license-over-role-in-russiagate-scandal",
    "https://www.bbc.com/news/articles/c20662dxl88o?at_medium=RSS&at_campaign=rss",
    "https://www.foxnews.com/us/second-federal-judge-blocks-justice-department-bid-release-jeffrey-epstein-grand-jury-files",
    "https://www.nbcnews.com/news/us-news/construction-companies-blamed-outbreak-legionnaires-rcna226104",
    "https://www.nbcnews.com/news/asian-america/fort-bliss-japanese-americans-internment-camp-immigrant-detention-rcna226044",
    "https://www.cbsnews.com/news/ernesto-barajas-killed-musician-drug-cartels-shot-dead-mexico/",
    "https://www.nbcnews.com/news/us-news/california-resident-catches-plague-likely-infected-flea-lake-tahoe-off-rcna226105",
    "https://www.cbsnews.com/news/abrupt-antarctic-climate-shifts-catastrophic-consequences-for-generations/",
    "https://www.foxnews.com/politics/fbi-arrests-woman-ten-most-wanted-fugitives-hiding-india-transports-us-prosecution",
    "https://www.foxnews.com/media/amanda-knox-how-one-misunderstood-text-sparked-suspicion-infamous-murder-case",
    "https://www.foxnews.com/sports/former-vikings-player-speaks-out-against-team-employing-male-cheerleaders",
    "https://www.foxnews.com/politics/illegal-trucker-deported-himself-ca-lawmaker-says-revealing-systemic-crisis-transportation",
    "https://www.bbc.com/news/articles/c939v4kw2l4o?at_medium=RSS&at_campaign=rss",
    "https://www.nbcnews.com/sports/tennis/carlos-alcaraz-emma-raducanu-us-open-first-round-doubles-rcna226089",
    "https://www.bbc.com/news/articles/cm2k1m56xgjo?at_medium=RSS&at_campaign=rss",
    "https://www.bbc.com/news/articles/cn92e52rpjxo?at_medium=RSS&at_campaign=rss",
    "https://www.foxnews.com/media/white-house-rips-ex-white-stripes-singer-slamming-trump-oval-office-redecorating-gaudy",
    "https://www.nbcnews.com/business/markets/trump-bought-bonds-companies-governments-rcna226024",
    "https://www.bbc.com/news/articles/c30z944q42do?at_medium=RSS&at_campaign=rss",
    "https://www.cbsnews.com/news/trump-doj-crime-victims-sanctuary-cities-lawsuit/",
    "https://www.foxnews.com/politics/desantis-state-line-wont-protect-criminal-illegals-sends-chuck-norris-deputy-nab-fugitive-trucker",
    "https://www.bbc.com/news/articles/cq587vpe46eo?at_medium=RSS&at_campaign=rss",
    "https://www.bbc.com/news/articles/c5yer12rp8po?at_medium=RSS&at_campaign=rss",
    "https://www.foxnews.com/media/chuck-todd-argues-putin-made-huge-miscalculation-not-giving-trump-win-alaska-summit",
    "https://www.nbcnews.com/select/shopping/best-hot-sauces-tested-rcna206590",
    "https://www.nbcnews.com/select/shopping/new-notable-august-2025-rcna225665",
    "https://www.nbcnews.com/pop-culture/pop-culture-news/justin-bieber-impersonator-dupes-dj-ends-las-vegas-stage-rcna226041",
    "https://www.foxnews.com/politics/illegal-migrants-charged-rape-aggravated-sodomy-kidnapping-woman-kansas-ice",
    "https://www.nbcnews.com/tech/misinformation/doctors-say-medical-misinformation-gotten-worse-survey-finds-rcna225804",
    "https://www.cbsnews.com/news/us-tariffs-trump-inflation-prices/",
    "https://www.cbsnews.com/losangeles/news/menendez-brothers-california-parole-board-hearings/",
    "https://www.bbc.com/weather/articles/cwy5r7xwq8xo?at_medium=RSS&at_campaign=rss",
    "https://www.cbsnews.com/news/us-anti-american-views-immigration-benefits/",
    "https://www.nbcnews.com/politics/justice-department/judge-denies-doj-request-release-jeffrey-epstein-grand-jury-transcript-rcna226117",
    "https://www.nbcnews.com/weather/heat/dangerous-heat-wildfire-conditions-take-hold-west-rcna226067",
    "https://www.cbsnews.com/news/this-week-on-sunday-morning-august-17-2025/",
    "https://www.bbc.com/sport/rugby-union/articles/c9d0nwepedeo?at_medium=RSS&at_campaign=rss",
    "https://www.cbsnews.com/news/jeffrey-epstein-grand-jury-judge/",
    "https://www.bbc.com/sport/tennis/articles/c5ypke1n361o?at_medium=RSS&at_campaign=rss",
    "https://www.cbsnews.com/news/powerball-jackpot-numbers-lottery-drawing-643-million/",
    "https://www.cbsnews.com/news/texas-house-special-session-redistricting-vote-today/",
    "https://www.bbc.com/news/articles/cj6yd4d5zj6o?at_medium=RSS&at_campaign=rss",
    "https://www.nbcnews.com/news/us-news/man-accused-stalking-family-unitedhealthcare-ceo-put-leave-rcna226028",
    "https://www.nbcnews.com/news/us-news/construction-companies-blamed-outbreak-legionnaires-rcna226104",
    "https://www.foxnews.com/health/five-deaths-reported-amid-bacterial-infection-outbreak-major-city",
    "https://www.cbsnews.com/news/trump-putin-zelenskyy-meeting-russia-ukraine/",
    "https://www.cbsnews.com/news/covid-19-shot-recommendations-american-academy-of-pediatrics-cdc/",
    "https://www.bbc.com/news/articles/cyvn9z3y78lo?at_medium=RSS&at_campaign=rss",
    "https://www.cbsnews.com/news/ps5-price-increase-sony/",
    "https://www.foxnews.com/politics/bostons-wu-fires-back-bondi-other-cities-slam-feds-over-sanctuary-warnings",
    "https://www.msnbc.com/top-stories/latest/joseph-humire-defense-department-pentagon-tren-de-aragua-rcna225929",
    "https://www.msnbc.com/top-stories/latest/texas-democrat-spends-night-house-floor-nicole-collier-rcna225907",
    "https://www.msnbc.com/top-stories/latest/democrats-senate-bid-sherrod-brown-ohio-jd-vance-rcna225866",
    "https://www.bbc.com/news/articles/c4gmjm8z47jo",
    "https://www.bbc.com/news/articles/cdrk267ldy7o",
    "https://www.bbc.com/news/articles/cm2vz9jg8jzo",
    "https://www.nbcnews.com/business/markets/nasdaq-tumbles-jackson-hole-jitters-hit-tech-stocks-rcna225890",
    "https://www.nbcnews.com/news/us-news/fake-labubu-dolls-knows-lafufus-are-serious-choking-hazards-kill-kids-rcna225691",
    "https://www.nbcnews.com/business/markets/nasdaq-tumbles-jackson-hole-jitters-hit-tech-stocks-rcna225890",
    "https://www.nbcnews.com/news/us-news/walmart-shrimp-may-exposed-radioactive-material-fda-says-rcna225903",
    "https://www.cbsnews.com/news/venezuela-nicolas-maduro-deploying-militia-us-threats/",
]


val_urls = [
    "https://www.foxnews.com/health/bindi-irwin-surgery-valley-fever-spike-fat-burning-treadmill-trend",
    "https://www.bbc.com/news/articles/cn85z0z3gz3o?at_medium=RSS&at_campaign=rss",
    "https://www.foxnews.com/travel/hurricane-erin-prompts-beach-closures-emergency-declarations-across-multiple-states",
    "https://www.foxnews.com/politics/full-s-new-york-republican-accuses-state-dems-hypocrisy-redistricting-push",
    "https://www.bbc.com/news/articles/c5y3599gx4qo",
    "https://www.cbsnews.com/news/politics-may-have-spurred-august-purge-of-5-veteran-fbi-agents/",
    "https://www.foxnews.com/politics/schiff-launches-legal-defense-fund-response-claims-trump-weaponizing-justice-system",
    "https://www.msnbc.com/rachel-maddow-show/maddowblog/trump-rig-midterm-elections-redistricting-mail-in-voting-census-rcna225826",
]

train_scenarios = [FactsScenario(article_url=url) for url in train_urls]

val_scenarios = [FactsScenario(article_url=url) for url in val_urls]
